//
// Communication Contention Model
//
// evaleng.cpp - Implementation of evaluation engine class
//
// Stathis Papaefstathiou (efp@microsoft.com)
// Copyright (c) 2000 Microsoft Research Ltd. All rights reserved.
//

#include <iostream>
#include <iomanip>
#include <sstream>
#include <limits>
#include <math.h>
#include <assert.h>
#include <stdexcept>
#include "ccmod.h"

// Comment in turn off otrace object references
// and optimize evaluation execution
#define USE_OTRACE      

using namespace std;

evaleng::evaleng(workload* w,netsys* s,otrace* t)
{
    nproc = s->GetNproc();
    linkno = s->GetLinkNo();
    wrkld = w;
    sys   = s;
    otr   = t;
    clock = 0;

    // Create processor objects
    p = new proc[nproc];

    for(int  i = 0; i < nproc; i++ )
	    p[i].SetId(i);

    // Create contention vector
    
    Vc = new float[linkno][2];
}

evaleng::~evaleng()
{
    delete p;
    delete [] Vc;
}

long evaleng::Go(void)
{
    long event = 1;        // Start event cycle

#ifdef USE_OTRACE
    otr->RecvSignal(otrace::CONFIG,nproc,linkno,wrkld,p);
    
    otr->RecvSignal(otrace::GO_INIT,sys->Name());
#endif

    while( CreateEvent() != 0 ) {
    	SysCont();
	    MsgCont();
	    CommCost();
	    EventHoriz();
	    UpdateProc();

    	event++;

#ifdef USE_OTRACE
        otr->RecvSignal(otrace::GO_STARTEVENT,event);
#endif
    }

#ifdef USE_OTRACE
        otr->RecvSignal(otrace::GO_END,event);
#endif

    return clock;
}


// Create new event for each proc
int evaleng::CreateEvent(void)
{
    // Message counts
    int newev = 0,
	oldev = 0;

    // For each processor create new event
    for( int i = 0; i < nproc; i++ ) {

    	// Already in event
    	if( p[i].isbusy() ) {
            if( p[i].GetEvent() == proc::PROC || p[i].issource() )
    		    oldev++;
    	    continue;
    	}

    	// No more events - processor exhausted event list
        if( wrkld->GetTraceType(i) == workload::TRACE_END )
    	    continue;

        // Trace extensions used as workload output
        if( wrkld->GetTraceType(i) > workload::TRACE_END )
            throw ccmod_error(__LINE__,"evaleng:fatal","Invalid trace type detected");

        // Increase new event counter
        if( wrkld->GetTraceType(i) == workload::TRACE_SCOM )
            newev += NewScom(i);
        else // Idle proc event
            newev += NewProc(i);
    }

#ifdef USE_OTRACE
        otr->RecvSignal(otrace::CREATEEVENT,newev,oldev);
#endif

    return newev+oldev;
}

// NewScom - Add a new sync comm event
int evaleng::NewScom(int i)
{
    // Message tuple
    int sp,tp;
    long ml;

    // Obtain current comm
    wrkld->GetTraceData(i,sp,tp,ml);

    // Check if processor is source
    msg cpmsg(sp,tp,ml);

    if( cpmsg.GetSrc() != i )
        return(0);  // no comm established

   	// Check is target is busy
    if( p[cpmsg.GetTrg()].isbusy() )
        return(0);

    // Obtain target message info
    wrkld->GetTraceData(cpmsg.GetTrg(),sp,tp,ml);
    msg tmsg(sp,tp,ml);

  	// Check if message is in both proc queues tops
    if( cpmsg != tmsg )
        return(0);

   	// Create new active message
    p[i].SetEvent(proc::SCOM);
    p[i].MsgActivate(cpmsg,sys);
    p[cpmsg.GetTrg()].SetEvent(proc::SRCV);

    // Synchronise clocks
    long cldf = p[i].GetClock()-p[cpmsg.GetTrg()].GetClock();
    
    p[i].SetClock( cldf < 0 ? -cldf : 0.0);
    p[i].SetTimeType(proc::WAIT, cldf < 0 ? -cldf : 0.0);
    p[cpmsg.GetTrg()].SetClock( cldf > 0 ? cldf : 0.0 );
    p[cpmsg.GetTrg()].SetTimeType( proc::WAIT, cldf > 0 ? cldf : 0.0 );


   	// Obtain next message from workload queues
    wrkld->FetchNextTrace(i);
    wrkld->FetchNextTrace(cpmsg.GetTrg());

#ifdef USE_OTRACE
    otr->RecvSignal(otrace::NEWSCOM,i);
#endif

    return(1); // add new event to global counter
}

// Add a new proc idle event
int evaleng::NewProc(int i)
{
    long t;

    // Add event into proc status
    p[i].SetEvent(proc::PROC);
    wrkld->GetTraceData(i,t);
    p[i].SetEventClock(t);

    // Prepare next trace
    wrkld->FetchNextTrace(i);

#ifdef USE_OTRACE
    otr->RecvSignal(otrace::NEWPROC,i,t);
#endif

    return(1); // add new event to global counter always
}

// Calc contention vector from active messages
void evaleng::SysCont(void)
{
    unsigned char (*Vr)[2];
	int i,j;

    // Initialise from previous use
    for( i = 0; i < linkno; i++ )
	    Vc[i][0] = Vc[i][1] = 0.0;

    // Calc link traffic
    for( i = 0; i < nproc; i++ ) {
        if( !p[i].issource() )
	        continue;
	    Vr = p[i].GetActMsg()->GetVr();

	    for( j = 0; j < linkno; j++ ) {
	        Vc[j][0] += Vr[j][0];
	        Vc[j][1] += Vr[j][1];
	    }
    }

    // Complete contention vector by including background loads
    float (*Vbgr)[2] = sys->GetBgrLoad(clock);
    for( i = 0; i < linkno; i++ ) {
        Vc[i][0] = Vc[i][0] == 0.0 ? 0.0 : (1-Vbgr[i][0]) / Vc[i][0];
        Vc[i][1] = Vc[i][1] == 0.0 ? 0.0 : (1-Vbgr[i][1]) / Vc[i][1];
    }

#ifdef USE_OTRACE
    otr->RecvSignal(otrace::SYSCONT,Vc);
#endif
}

// Calculate message contention
void evaleng::MsgCont(void)
{
    unsigned char (*Vr)[2];
    float maxcont[2];

    // For each proc
    for( int i = 0; i < nproc; i++ ) {
    	if( !p[i].issource() )
	        continue;

    	Vr = p[i].GetActMsg()->GetVr();
	    maxcont[0] = maxcont[1] = 0.0;

    	// Calc message contention
	    for( int j = 0; j < linkno; j++ ) {
	        if( Vr[j][0] != 0 && (maxcont[0] == 0.0 || maxcont[0] > Vc[j][0]) )
		        maxcont[0] = Vc[j][0];
	        if( Vr[j][1] != 0 && (maxcont[1] == 0.0 || maxcont[1] > Vc[j][1]) )
		        maxcont[1] = Vc[j][1];
	    }

        assert(maxcont[0] != 0.0 || maxcont[1] != 0.0 );

    	// Set message contention
	    p[i].GetActMsg()->SetCont(maxcont);

#ifdef USE_OTRACE
        otr->RecvSignal(otrace::MSGCONT,i,maxcont);
#endif
    }
}

// Calculate cost for communication
void evaleng::CommCost(void)
{
    // For each processor
    for(int i =0; i < nproc; i++) {
    	// For each active message
	    if( !p[i].issource() )
	        continue;

    	// Get number of packets
	    float npack = p[i].GetActMsg()->GetPacks();

       	// Get number of hops for message
    	int hops = p[i].GetActMsg()->GetHops();

    	// Calc communication cost & add to proc
        long ccost = sys->Tcom(hops,npack)/p[i].GetActMsg()->GetCont();

	    p[i].SetEventClock(ccost);

#ifdef USE_OTRACE
        otr->RecvSignal(otrace::COMMCOST,i,npack,hops,ccost);
#endif

    }
}

// Calculate event horizon
void evaleng::EventHoriz(void)
{
    evhor = numeric_limits<unsigned long>::max(); // flag non setting

    for( int i=0 ; i < nproc; i++ ) {
        if( p[i].GetEvent() != proc::PROC && p[i].GetEvent() != proc::SCOM )
	        continue;

        // First non zero value
        if( evhor == numeric_limits<unsigned long>::max() && p[i].GetEventClock() >= 0 )
            evhor = p[i].GetEventClock();

        // Found shorter event
        if( evhor > p[i].GetEventClock() )
            evhor = p[i].GetEventClock();
    }

    // Internal error if event horizon is 0
    if( evhor == numeric_limits<unsigned long>::max() )
        throw ccmod_error(__LINE__,"internal","No active events detected");

#ifdef USE_OTRACE
        otr->RecvSignal(otrace::EVENTHORIZON,evhor);
#endif

}

// Update proc status
void evaleng::UpdateProc(void)
{

    for( int i=0; i < nproc; i++ )  {

        switch(p[i].GetEvent()) {
        case proc::PROC:

            p[i].SetClock(evhor);
            p[i].SetTimeType(proc::PROC,evhor);
            p[i].SetEventClock(p[i].GetEventClock()-evhor);

            // end of proc event
            if( p[i].GetEventClock() == 0 ) {
                p[i].SetNotActive();

#ifdef USE_OTRACE
                otr->RecvSignal(otrace::UPPR_PROC,i);
#endif
            }

            break;
        case proc::SCOM:
            actmsg* m = p[i].GetActMsg();
            float packs = m->RoundPack(sys->T2Pack(evhor,m->GetHops(),
                m->GetTotalPacks(),m->GetPacks(),m->GetTcom()) *m->GetCont());

            assert(packs >= 0.0);

            p[i].SetClock(evhor);
            p[i].SetTimeType(proc::SCOM,evhor);
            m->TcomCommit(evhor);
            m->PackCommit(packs);
            p[i].SetEventClock(p[i].GetEventClock()-evhor);

#ifdef USE_OTRACE
            int ptrg = m->GetSmsg().GetTrg();

            otr->RecvSignal(otrace::UPPR_PCNS,i,ptrg,packs);
#endif

        	if( ! m->isfinished() )
	            continue;

            // Update src & target procs
	        int trg = m->GetSmsg().GetTrg();
	        p[trg].SetNotActive();
	        p[trg].SetClock(p[i].GetActMsg()->GetTcom());
            p[trg].SetTimeType(proc::SRCV,p[i].GetActMsg()->GetTcom());
	        p[i].MsgComplete();

#ifdef USE_OTRACE
            otr->RecvSignal(otrace::UPPR_SCOM,i,trg);
#endif
            break;
        }

        // Update evaleng clock
        if( p[i].GetClock() > clock )
            clock = p[i].GetClock();
    }

#ifdef USE_OTRACE
    otr->RecvSignal(otrace::UPPR_END,clock);
#endif

}

