//
// Communication Contention Model
//
// myrinet.cpp - Implementation of myrinet class
//
// NETSYS DERIVED CLASS
// MYRINET M2M-OCT-SW8 Interconection Network
//
// Stathis Papaefstathiou (efp@microsoft.com)
// Copyright (c) 2000 Microsoft Research Ltd. All rights reserved.
//

#include <iostream>
#include <sstream>
#include <assert.h>
#include <stdlib.h>

#include "ccmod.h"
#include "myrinet.h"

using namespace std;

/*
 * MYRINET Switch M2M-OCT-SW8
 * Multistage toplogy
 * Static Routing
 * Packet size is 20 bytes header + 2028 bytes data
 * Supports var size packets. 
 * HPVM 1.9, MPI 1.1 library running on Win 2000
 */

static const int  _SwitchLinks = 32;    // Fixed for 1 switch
static const int  _Psize = 2028;        // Data packet size in bytes (+20 bytes header)
static const char _SysNam[] = "Myrinet M2M-OCT-SW8, NT 4.0, HPVM 1.9, MPI 1.1";
static int  _S1toS2[16]; // 1st level switch -> 2nd level switch mapping

myrinet::myrinet(int p)
{
    Nproc = 0;
    Vbgr = NULL;

    // Determine static S1->S2 level connections
    for( int i = 0; i < 16; i++ )
        _S1toS2[i] = (int)ceil(((float)rand()/RAND_MAX) * 4);

    if( p < 1 || p > 16 )
        throw ccmod_error(__LINE__,"Myrinet","Processors should be 1-16");

    Nproc = p;

    Nlink = _SwitchLinks;

    if( Vbgr != NULL )
        delete [] Vbgr;

    Vbgr = new float[Nlink][2];
    for(i = 0; i < Nlink; i++)
        Vbgr[i][0] = Vbgr[i][1] = 0.0;
}


myrinet::~myrinet()
{
    delete [] Vbgr;
}


int myrinet::GetLinkNo(void)
{
    return Nlink;
}

const char* myrinet::Name(void)
{
    return _SysNam;
}

int myrinet::GetNproc(void)
{
    return Nproc;
}

// Calculate communication cost
long myrinet::Tcom(int hops, float npack)
{
    // From experiments hops do not change
    // substancialy the regression model

    return npack <= 1.0 ? (14.68 + 0.0334 * _Psize * npack) : (54.87 + 0.0143 * _Psize *npack) ;
}

// Calculate packs completed in specified period
float myrinet::T2Pack(long tevent,int hops,float tpacks,float rpacks, long tcom)
{
    float tstart = tpacks <= 1.0 ? 14.68 : 54.87;
    float coef   = tpacks <= 1.0 ? 0.0334 : 0.0143;

    if( tcom < tstart )
        tevent -= tstart-tcom < tevent ? tstart-tcom : tevent;

    return (tevent > 0.0 ?  (tevent / (coef*_Psize)) : 0.0 );
}

// Return packet size
int myrinet::PacketSize(void)
{
    return _Psize;
}

// Main routing algorithm
void myrinet::Routing(int Psrc,int Pdst, unsigned char (*Vr)[2])
{
    int SrcGr = Psrc/4;     // Src proc group
    int TrgGr = Pdst/4;     // Trg proc group

    // Initialise vector
    for( int i = 0; i < Nlink; i++ )
        Vr[i][0] = Vr[i][1] = 0;

    //
    // Routing
    //

    Vr[Psrc][0] = 1;    // Psrc -> 1st level switch

    // 
    // If Ptrg && Pdst don't belong to same group
    // involve 2nd level switch
    //
    if( SrcGr != TrgGr ) {
        // 
        // Use static preselected routing to select 2nd level switch
        //
        int selsw = _S1toS2[Psrc];

        // Check validity of link #
        assert(SrcGr*4 + selsw + 15 > 15 && SrcGr*4 + selsw + 15 < 32);

        Vr[ SrcGr*4 + selsw + 15][0] = 1;

        //
        // Set busy link from 2nd level switch
        // to target proc 1st level switch
        //

        // Check validity of link #
        assert(selsw + TrgGr * 4 + 15 > 15 && selsw + TrgGr * 4 + 15 < 32);

        Vr[selsw + TrgGr * 4 + 15][1] = 1;
    }
        
    Vr[Pdst][1] = 1;    // 2nd level switch -> Ptrg
}

// Return background load vector
float (*myrinet::GetBgrLoad(long clock))[2]
{
    // No background load specified
    return(Vbgr);
}


// ---------------------------------------------------------------------

