//
// Communication Contention Model
//
// wrkascii.cpp - Implementation of wrkascii class
//
// WORKLOAD DERIVED CLASS
// Read & process ASCII workload traces
//
// Stathis Papaefstathiou (efp@microsoft.com)
// Copyright (c) 2000 Microsoft Research Ltd. All rights reserved.
//

#include <iostream>
#include <sstream>

#include "ccmod.h"

#include "wrkascii.h"

using namespace std;

//
// Additional traces used within PACE ASCII traces
//
#define TRACE_MARK      (workload::TRACE_EXT1)
#define TRACE_LOOP      (workload::TRACE_EXT2)

// Create & Destroy object

wrkascii::wrkascii(void)
{
    // Do nothing
    isconfig = false;
}

wrkascii::wrkascii(int proc,int iproc,istream* infl)
{
    isconfig = false;
    Config(proc,iproc,infl);
}


wrkascii::~wrkascii(void)
{
    for( int i = 0; i < nproc; i++ )
        pdata[i].ptraces.clear();

	delete[] pdata;
}

// Generic config function
void wrkascii::Config(int proc,int iproc,istream* infl)
{
    int  i;
    char c;
    ostringstream os;

    // Make sure this is a first
    if( isconfig)
         throw ccmod_error(__LINE__,"ASCII traces","Reconfiguration is invalid");

	// Setup configuration
	nproc = proc;
    isconfig = true;

	// Setup processor workload lists
    pdata = new processor[nproc];
        
	// Read trace file and store it to processor trace queues

	long line = 1;   // Current trace line
	asciitrace curt; // Current trace
	char cmd;		 // Trace command

	while( 1 ) {
    	char d1,d2;
        int  p;

		// Read trace type
		while( (isspace(cmd = infl->get()) || cmd == '\n') && infl->good() )
            ;

		if( infl->eof() || infl->bad())
			break;

		// Convert to uppercase
		cmd = toupper(cmd);

		switch(cmd) {
		case '#':	// comment
            // Do nothing
			break;
		case 'C':	// sync comm
      
			curt.tt = TRACE_SCOM;

			*infl >> curt.src >> d1 >> curt.trg >> d2 >> curt.len;

			curt.src -= iproc;
			curt.trg -= iproc;

			// Check format of trace
			if( !infl->good() || d1 != ',' || d2 != ',' ) {
				os << "Invalid sync comm trace, trace line " << line;

                throw ccmod_error(__LINE__,"ASCII traces",os.str());
			}
			
			// Is entry valid ?
			if( curt.src < 0 || curt.trg < 0 || curt.src > nproc-1 || 
				curt.trg > nproc-1 || curt.len < 0 || curt.src == curt.trg ) {

				os << "Invalid arguments, trace line " << line;

                throw ccmod_error(__LINE__,"ASCII traces",os.str());
			}

            // Insert comm trace to source & target proc queues
			pdata[curt.src].ptraces.push_back(curt);
			pdata[curt.trg].ptraces.push_back(curt);

			break;
		case 'P':	// Processing

            curt.tt = workload::TRACE_IDLE;

			*infl >> p >> d1 >> curt.time;

            p -= iproc;

			// Check format of trace
			if( !infl->good() || d1 != ',' ) {
				os << "Invalid processing, trace line " << line;

                throw ccmod_error(__LINE__,"ASCII traces",os.str());
			}

			// Is entry valid ?
			if( p < 0 || p > nproc-1 || curt.time < 0 ) {
				os << "Invalid arguments, trace line " << line;

                throw ccmod_error(__LINE__,"ASCII traces",os.str());
			}

            // Insert trace to processor queue
			pdata[p].ptraces.push_back(curt);

			break;
		default:
			os << "Uknown trace type, trace line " << line;

            throw ccmod_error(__LINE__,"ASCII traces",os.str());
			break;
		}

        // Clear anything after trace
		while( (c = infl->get()) != EOF && c != '\n' )
			;

		line++;
	}

 
    // Insert TRACE_END and find first trace/proc
    curt.tt = workload::TRACE_END;

    for( i = 0; i < nproc; i++ ) {
        pdata[i].ptraces.push_back(curt);
        _NextTrace(true,i);
    }

}

// ----------------------------------------------------------------------

/*
 * Trace Access Functions
 */

// GetTraceType -- Return current trace type
workload::trace wrkascii::GetTraceType(int procid)
{
	return pdata[procid].curtrace->tt;
}

// GetTraceComm - Return comm trace arguments
void wrkascii::GetTraceData(int procid,int &src,int &trg,long &len)
{
    src = pdata[procid].curtrace->src;
    trg = pdata[procid].curtrace->trg;
    len = pdata[procid].curtrace->len;
}

// GetTraceProc -- Return processing trace arguments
void wrkascii::GetTraceData(int procid, long &time)
{
    time = pdata[procid].curtrace->time;
}

// FetchNextTrace -- Return next trace
workload::trace wrkascii::FetchNextTrace(int procid)
{
    _NextTrace(false,procid);

    return GetTraceType(procid);
}

// ----------------------------------------------------------------------

/* 
 * Output status functions
 */
void wrkascii::print(ostream &os)
{
    int empty = 0;

    os << "Nproc " << nproc << ", obj pointer " << this << endl;

    for( int i=0; i < nproc; i++ ) {

        if( pdata[i].curtrace->tt == TRACE_END ) {
            empty++;
            continue;
        }

        os << "Proc: " << i << ", Top Trace: ";

        switch(pdata[i].curtrace->tt) {
        case workload::TRACE_IDLE:
            os << "process (" << pdata[i].curtrace->time << ") ";
            break;
        case workload::TRACE_SCOM:
            os << "sync comm (" << pdata[i].curtrace->src << "," <<
                pdata[i].curtrace->trg << "," << pdata[i].curtrace->len << ") ";
            break;
        default:
            os << "Control trace type " << (int)pdata[i].curtrace->tt;
        }

        os << endl;
    }

    if( empty == nproc )
        os << "Processor workload queues are empty" << endl;
}


// ----------------------------------------------------------------------
 
/*
 * Private Trace Access Methods
 */

// _NextTrace -- Goes through the list to find next trace. 
void wrkascii::_NextTrace(bool isfirst, int proc)
{
    // Initialise processor struct in first call
    if( isfirst )
        pdata[proc].curtrace = pdata[proc].ptraces.begin();

    // Check for end of traces
    if( pdata[proc].curtrace->tt == TRACE_END )
        return; // Do nothing

    // Next trace
    if( !isfirst ) {
        pdata[proc].curtrace++;

        // Delete old previous trace
        list<asciitrace>::iterator prv;
        prv = pdata[proc].curtrace;
        pdata[proc].ptraces.erase(--prv);
    }
}
